/*
	Copyright (c) 1992
	Digital Equipment Corporation, Maynard, Massachusetts 01754

	The HAL Alpha Assembler is free software; you can redistribute 
	it and/or modify it with the sole restriction that this copyright 
	notice be attached.

	The information in this software is subject to change without notice 
	and should not be construed as a commitment by Digital Equipment
	Corporation.

	The author disclaims all warranties with regard to this software,
	including all implied warranties of merchantability and fitness, in
	no event shall he be liable for any special, indirect or consequential
	damages or any damages whatsoever resulting from loss of use, data, 
	or profits, whether in an action of contract, negligence or other
	tortious action, arising out of or in connection with the use or
	performance of this software.

*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "pp.h"

#if defined(VMS) || defined(__MSDOS__)
#include "pp_tab.h"
#else
#include "pp.tab.h"
#endif

extern int yychar;
extern YYSTYPE yylval;

static char *ident_chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_$";

#ifdef __STDC__
void do_include(TOKEN_LIST *name_tok)
#else
void do_include(name_tok)
TOKEN_LIST *name_tok;
#endif
{
  char *name = name_tok->tok.s;
  INPUT_FILE *temp;

  /* push cur_ifile back on ifile stack */
  cur_ifile->yychar = yychar;
  cur_ifile->yylval = yylval;
  cur_ifile->next = ifile;
  ifile = cur_ifile;

  /* create INPUT_FILE struct for this include file on ifile stack */
  add_input(name, F_FILE, 1, ADD_PUSH);

  /* call next_file() to get input file */
  (void)next_file();

  free_tok(name_tok);

  return;
}

#define COND_EQ(t)	(((t)->tok.q.v[0] == 0) && ((t)->tok.q.v[1] == 0))
#define COND_GT(t)	((((t)->tok.q.v[1] & 0x80000000) == 0) && (((t)->tok.q.v[0] != 0) || ((t)->tok.q.v[1] != 0)))
#define COND_LT(t)	((((t)->tok.q.v[1] & 0x80000000) != 0) && (((t)->tok.q.v[0] != 0) || ((t)->tok.q.v[1] != 0)))

#ifdef __STDC__
void do_if(TOKEN_LIST *cond, TOKEN_LIST *exp, TOKEN_LIST *body, TOKEN_LIST *endbody)
#else
void do_if(cond, exp, body, endbody)
TOKEN_LIST *cond, *exp, *body, *endbody;
#endif
{
  TOKEN_LIST *result;
  int is_true = 0;
  int pushed_curifile = 0;

  /* if a numeric condition, evaluate expression */
  if ((cond->type != B) && (cond->type != NB)) {
    result = eval_const_expr(exp);
    if ((result != NULL) && (result->type != NUMBER)) {
      errout(E_ERR, "do_if", "expression does not evaluate to a constant.\n");
      return;
    }
  }

  switch (cond->type) {
  case EQ:
    is_true = (result != NULL) && COND_EQ(result);
    break;
  case GT:
    is_true = (result != NULL) && COND_GT(result);
    break;
  case LT:
    is_true = (result != NULL) && COND_LT(result);
    break;
  case NE:
    is_true = (result != NULL) && !COND_EQ(result);
    break;
  case GE:
    is_true = (result != NULL) && (COND_EQ(result) || COND_GT(result));
    break;
  case LE:
    is_true = (result != NULL) && (COND_EQ(result) || COND_LT(result));
    break;
  case DF:
    is_true = result != NULL;
    break;
  case NDF:
    is_true = result == NULL;
    break;
  case B:
    /* argument was blank if macro argument replacement resulted in no ARG_VALUE, and
       therefore no expression */
    is_true = exp == NULL;
    break;
  case NB:
    /* argument was not blank if there was an ARG_VALUE specified */
    is_true = exp != NULL;
    break;
  }

  /* okay, we have the result, now process all the if clauses that are appropriate.

     NOTE: the string passed with the token is the body for the PREVIOUS token. This
     is a parsing artifact that is not worth working around in FLEX, so we deal with it here.

     E.G. take the following example:
     	.if eq 0
	    <code1>
	.iff
	    <code2>
	.endc

     FLEX will parse as follows: IF cond expression EOL IFF ENDC EOL
     if IFF token will contain <code1> which is executed if the condition was true,
     and the ENDC token will contain <code2> which is executed if the condition is
     false. */
  /* handle <code1> (see above) case first */
  if (is_true) {
    if (!pushed_curifile) {
      /* push cur_ifile back on ifile stack */
      cur_ifile->yychar = yychar;
      cur_ifile->yylval = yylval;
      cur_ifile->next = ifile;
      ifile = cur_ifile;
      
      /* set flag so we don't do this again. */
      pushed_curifile = 1;
    }
    
    /* add first portion of the if body */
    if (body != NULL)
      add_input(body->tok.s, F_STRING, 1, ADD_PUSH);
    else /* note that endbody can not be NULL. */
      add_input(endbody->tok.s, F_STRING, 1, ADD_PUSH);    
  }

  while (body != NULL) {
    int do_push;

    switch (body->type) {
    case IFF:
      do_push = !is_true;
      break;
    case IFT:
      do_push = is_true;
      break;
    case IFTF:
      do_push = 1;
      break;
    default:
      errout(E_ERR, "do_if", "illegal body token.\n");
      break;
    }

    if (do_push) {
      if (!pushed_curifile) {
	/* push cur_ifile back on ifile stack */
	cur_ifile->yychar = yychar;
	cur_ifile->yylval = yylval;
	cur_ifile->next = ifile;
	ifile = cur_ifile;

	/* set flag so we don't do this again. */
	pushed_curifile = 1;
      }
      
      /* add this portion of the if body */
      if (body->next != NULL)
	add_input(body->next->tok.s, F_STRING, 1, ADD_PUSH);
      else /* note that endbody can not be NULL. */
	add_input(endbody->tok.s, F_STRING, 1, ADD_PUSH);

    }
    body = body->next;
  }

  if (pushed_curifile) {
    /* call next_file() to get input file */
    (void)next_file();
  }

  free_tok(result);
  free_tok(cond);
  free_tok_list(body);
  free_tok(endbody);

  return;    
}

#ifdef __STDC__
void do_align(TOKEN_LIST *exp)
#else
void do_align(exp)
TOKEN_LIST *exp;
#endif
{
  TOKEN_LIST *result;
  TOKEN_LIST *exp1, *exp2;

  exp2 = exp1 = exp;
  /* use result as a temp variable */
  result = NULL;
  while ((exp2 != NULL) && (exp2->type != ',')) {
    result = exp2;   /* save previous */
    exp2 = exp2->next;
  }

  /* break end of first expression before ',' */
  if (result != NULL)
    result->next = NULL;

  /* skip by comma */
  if (exp2 != NULL)
    exp2 = exp2->next;

  /* check for quad or long keywords */
  if (exp1->type == SYMBOL) {
    if (strcmp(exp1->tok.s, "quad") == 0) {
      add_to_line("^x0000000000000003");
      goto do_exp2;
    }
    if (strcmp(exp1->tok.s, "long") == 0) {
      add_to_line("^x0000000000000002");
      goto do_exp2;
    }
  }    

  /* evaluate expression */
  result = eval_const_expr(exp1);
  
  if (result != NULL) {
    add_to_line(tok2str(result));
    free_tok(result);
  } else
    errout(E_ERR, "do_align", "alignment expression did not evaluate to a constant.\n");

do_exp2:
  if (exp2 != NULL) {
    /* evaluate second expression */
    result = eval_const_expr(exp2);
    if (result != NULL) {
      add_to_line(",");
      add_to_line(tok2str(result));
      free_tok(result);
    } else
      errout(E_ERR, "do_align", "fill expression did not evaluate to a constant.\n");
  } else
    add_to_line(",^x0000000000000000");

  return;
}

#ifdef __STDC__
void do_repeat(TOKEN_LIST *exp, TOKEN_LIST *body)
#else
void do_repeat(exp, body)
TOKEN_LIST *exp, *body;
#endif
{
  TOKEN_LIST *result = eval_const_expr(exp);

  if ((result != NULL) && (result->type != NUMBER)) {
    errout(E_ERR, "do_repeat", "expression does not evaluate to a constant.\n");
    return;
  }

  if (result->tok.q.v[1] != 0)
    errout(E_WARN, "do_repeat", "only 32 bit repeat counts allowed, upper bits assumed zero.\n");

  /* push cur_ifile back on ifile stack */
  cur_ifile->yychar = yychar;
  cur_ifile->yylval = yylval;
  cur_ifile->next = ifile;
  ifile = cur_ifile;

  add_input(body->tok.s, F_STRING, result->tok.q.v[0], ADD_PUSH);

  /* call next_file() to get input file */
  (void)next_file();

  free_tok(body);
  free_tok(result);

  return;    
}

#ifdef __STDC__
void do_mem_directive(TOKEN_LIST *which, TOKEN_LIST *explist, TOKEN_LIST *comment)
#else
void do_mem_directive(which, explist, comment)
TOKEN_LIST *which, *explist, *comment;
#endif
{
  TOKEN_LIST *nexpr;

  /* do once for each expression specified */
  nexpr = explist;
  while (nexpr != NULL) {
    TOKEN_LIST *stack = NULL, *temp;

    explist = nexpr;
    if (explist->next != NULL) {
      while (explist->next->type != ',') {
	explist = explist->next;
	if (explist->next == NULL)
	  break;
      }
      if (explist->next != NULL) {
	if(explist->next->next == NULL) {
	  errout(E_ERR, "do_mem_directive", "comma present without following expr in %s.\n", which->tok.s);
	  temp = NULL;
	} else
	  /* save next expression */
	  temp = explist->next->next;
      } else
	temp = NULL;
      /* stop this expression at the ',' */
      explist->next = NULL;
    } else
      temp = NULL;
    
    eval_expr(nexpr, &stack);

    output_assignments(defs);
    defs = NULL;

    add_to_line(which->tok.s);
    add_to_line(tok2str(stack));

    if ((nexpr = temp) == NULL)
      output_line(comment);
    else
      output_line(NULL);
  }

  free_tok(which);

  return;
}

#ifdef __STDC__
char *get_macro_arg(char *name)
#else
char *get_macro_arg(name)
char *name; 
#endif
{
  int len = strspn(name, ident_chars), i;
  char temp[40];
  
  for (i = 0; i < macstk->argn; ++i)
    if ((len == strlen(macstk->argl[i].def->tok.s)) &&
	(strncmp(name, macstk->argl[i].def->tok.s, len) == 0)) {
      /* it matches!!! */
      
      /* use actual argument if specified, otherwise use default */
      if (macstk->argl[i].act != NULL) {
	return macstk->argl[i].act->tok.s;
      } else {
	/* check for default argument */
	TOKEN_LIST *tok = macstk->argl[i].def->next;
	
	if (tok != NULL) {
	  switch (tok->type) {
	  case '=':
	    /* use a default value */
	    tok = tok->next;
	    return tok->tok.s;
	  case '?':
	    /* create a local label */
	    if (macstk->argl[i].labelnum == 0)
	      macstk->argl[i].labelnum = local_label++;
	    sprintf(temp, "%d$", macstk->argl[i].labelnum);
	    return temp;
	  default:
	    errout(E_ERR, "get_macro_arg", "illegal token type after argument name.\n");
	    break;
	  }
	} else
	  return "";
      }
      break;
    }

  return NULL;  
}

#ifdef __STDC__
TOKEN_LIST *get_symbol_value(char *name)
#else
TOKEN_LIST *get_symbol_value(name)
char *name; 
#endif
{
  SYMDEF *stemp;
  ASSIGN *atemp;
  TOKEN_LIST *result;

  if (strchr(name, '%') == NULL) {
    if (avl_lookup(symtree, name, &stemp)) {
      if (stemp->incarnation != S_LABEL) {
/*	modify_symbol(expr, stemp->incarnation - 1); */
	char junk[20], *temp;

	sprintf(junk, "%%%d", stemp->incarnation - 1);
	temp = name;
	name = ALLOC_STR(strlen(temp)+strlen(junk));
	strcpy(name, temp);
	strcat(name, junk);
      }
    } else
      return NULL;
  }

  if (avl_lookup(symtab, name, &atemp)) {
    TOKEN_LIST *texpr = copy_expr(atemp->operand1);
    result = eval_const_expr(texpr);
    if ((result != NULL) && (result->type != NUMBER)) {
      errout(E_ERR, "get_symbol_value", "expression does not evaluate to a constant.\n");
      return NULL;
    }
  } else {
    errout(E_ERR, "get_symbol_value", "symbol (%s) not found.\n", name);
    return NULL;
  }
  
  return result;
}

#ifdef __STDC__
char *extract_string(char *start, char *delim, char **value)
#else
char *extract_string(start, delim, value)
char *start, *delim, **value;
#endif
{
  /* NOTE: both extract_int and extract_string must NOT overwrite the final delimiter
     they find.  %locate looks one character backward to determine whether there is a 
     position argument or not.  */

  start += strspn(start, " \t");  /* clear out WS */
  if (*start == '<') {
    /* delimited string */
    *value = ++start;
    start = strchr(start, '>');
    if (start == NULL) {
      errout(E_ERR, "extract_string", "string macro argument did not have closing delimiter.\n");
      start = strpbrk(start, delim);
      if (start == NULL)
	return NULL;
      *(start++) = '\0';
      return start;
    }

    *(start++) = '\0';

    /* look for one of the delimiters, if none are there, we'll have to die */
    start = strpbrk(start, delim);
    if (start == NULL)
      return NULL;
    return start + 1;
  } else {
    /* macro arg */
    *value = get_macro_arg(start);

    /* look for one of the delimiters, if none are there, we'll have to die */
    start = strpbrk(start, delim);
    if (start == NULL)
      return NULL;
    return start + 1;
  }

  /* can't get here */
}

#ifdef __STDC__
char *extract_int(char *start, char *delim, int *value)
#else
char *extract_int(start, delim, value)
char *start, *delim;
int *value;
#endif
{
  char *ctmp;
  int temp;

  /* NOTE: both extract_int and extract_string must NOT overwrite the final delimiter
     they find.  %locate looks one character backward to determine whether there is a 
     position argument or not.  */

  start += strspn(start, " \t");  /* clear out WS */

  temp = strtol(start, NULL, 10);
  if ((temp == 0) && (*start != '0')) {
    /* symbol */
    TOKEN_LIST *result;
    char *symnam;
    
    /* zero terminate the symbol name */
    temp = strspn(start,ident_chars);
    symnam = strncpy(ALLOC_STR(temp), start, temp);
    symnam[temp] = '\0';
    ctmp = start + temp;

    result = get_symbol_value(symnam);
    
    if (result == NULL) {
      errout(E_ERR, "extract_int", "integer macro argument was not a constant.\n");
      *value = 0;
    } else {
      if (result->tok.q.v[1] != 0)
	errout(E_WARN, "prep_macro_line", "integer macro argument has non-zero value in upper 32 bits.\n");
      
      *value = result->tok.q.v[0];
    }
    start = strpbrk(ctmp, delim);
    return start + 1;
  } else {
    *value = temp;
    start = strpbrk(start, delim);
    if (start == NULL)
      return NULL;
    return start + 1;
  }
    
  /* can't get here */
}

#ifdef __STDC__
void prep_macro_line(char *body, char **newbody)
#else
void prep_macro_line(body, newbody)
char *body, **newbody;
#endif
{
  char *ctmp = macro_temp;
  int i;

  while (*body) {
    if (isalnum(*body) || (*body == '_') || (*body == '$')) {
      /* could be an argument */
      int len = strspn(body, ident_chars);
      
      for (i = 0; i < macstk->argn; ++i)
	if ((len == strlen(macstk->argl[i].def->tok.s)) &&
	    (strncmp(body, macstk->argl[i].def->tok.s, len) == 0)) {
	  /* it matches!!! it is a macro argument */
	  
	  /* check for preceding argument concatenator, if there, lose it */
	  if (*(ctmp-1) == '\'')
	    ctmp--;
	  /* use actual argument if specified, otherwise use default */
	  if (macstk->argl[i].act != NULL) {
	    strcpy(ctmp, macstk->argl[i].act->tok.s);
	    ctmp += strlen(macstk->argl[i].act->tok.s);
	  } else {
	    /* check for default argument */
	    TOKEN_LIST *tok = macstk->argl[i].def->next;
	    
	    if (tok != NULL) {
	      switch (tok->type) {
	      case '=':
		/* use a default value */
		tok = tok->next;
		strcpy(ctmp, tok->tok.s);
		ctmp += strlen(tok->tok.s);
		break;
	      case '?':
		/* create a local label */
		if (macstk->argl[i].labelnum == 0)
		  macstk->argl[i].labelnum = local_label++;
		sprintf(ctmp, "%d$", macstk->argl[i].labelnum);
		ctmp += strlen(ctmp);
		break;
	      default:
		break;
	      }
	    }
	  }
	  break;
	}
      if (i == macstk->argn) {
	strncpy(ctmp, body, len);
	ctmp += len;
      }
      body += len;
      /* if we replaced an argument, and there is an argument concatenation operator
	 ("'") after the argument, skip it. */
      if ((i != macstk->argn) && (*body == '\''))
	body++;
    } else
      if (*body == '%') {
	char tstr[10];
	
	strncpy(tstr, body, 9);
	for (i = 0; i < 9; ++i)
	  tstr[i] = isalpha(tstr[i])?tolower(tstr[i]):tstr[i];
	if (strncmp(tstr, "%length", 7) == 0) {
	  char *retval;
	  int len;

	  body = strchr(body, '(');
	  if (body == NULL) {
	    errout(E_ERR, "prep_macro_line", "no open paren for %length.\n");
	    *newbody = "";
	    return;
	  } else {
	    body = extract_string(++body, ")", &retval);
	    len = strlen(retval);
	    sprintf(ctmp, "%d", len);
	    ctmp += strlen(ctmp);
	  }
	} else
	  if (strncmp(tstr, "%locate", 7) == 0) {
	    char *s1, *s2;
	    int pos;

	    body = strchr(body, '(');
	    if (body == NULL) {
	      errout(E_ERR, "prep_macro_line", "no open paren for %locate.\n");
	      *newbody = "";
	      return;
	    } else {
	      char *cpos;

	      body = extract_string(++body, ",", &s1);
	      if (body == NULL) {
		errout(E_ERR, "prep_macro_line", "unable to parse substring argument in %locate.\n");
		*newbody = "";
		return;
	      }
	      body = extract_string(body, ",)", &s2);
	      if (body == NULL) {
		errout(E_ERR, "prep_macro_line", "unable to parse search string argument in %locate.\n");
		*newbody = "";
		return;
	      }
	      if (*(body-1) == ',') {
		body = extract_int(body, ")", &pos);
		if (body == NULL) {
		  errout(E_ERR, "prep_macro_line", "unable to parse position argument in %locate.\n");
		  *newbody = "";
		  return;
		}
	      } else
		pos = 0;

	      /* okay s1, s2, and pos are set up */
	      s2 += pos;
	      cpos = strstr(s2, s1);
	      if (cpos == NULL)
		sprintf(ctmp, "%d", strlen(s2) + pos);
	      else
		sprintf(ctmp, "%d", cpos - s2 + pos);
	      ctmp += strlen(ctmp);
	    }
	  } else
	    if (strncmp(tstr, "%extract", 8) == 0) {
	      char *str;
	      int sym1, sym2, i;

	      body = strchr(body, '(');
	      if (body == NULL) {
		errout(E_ERR, "prep_macro_line", "no open paren for %extract.\n");
		*newbody = "";
		return;
	      } else {
		body = extract_int(++body, ",", &sym1);
		if (body == NULL) {
		  errout(E_ERR, "prep_macro_line", "unable to position argument in %extract.\n");
		  *newbody = "";
		  return;
		}
		body = extract_int(body, ",", &sym2);
		if (body == NULL) {
		  errout(E_ERR, "prep_macro_line", "unable to parse length argument in %extract.\n");
		  *newbody = "";
		  return;
		}
		body = extract_string(body, ")", &str);
		if (body == NULL) {
		  errout(E_ERR, "prep_macro_line", "unable to parse string argument in %extract.\n");
		  *newbody = "";
		  return;
		}
		str += sym1;
		for (i = 0; (i < sym2) && (*str != '\0'); ++i)
		  *(ctmp++) = *(str++);
	      }
	    }
      } else 
	*(ctmp++) = *(body++);
  }
  *ctmp = '\0';

  *newbody = strcpy(ALLOC_STR(strlen(macro_temp)), macro_temp);
  
  return;
}

#ifdef __STDC__
void free_macro_args(MACRO_STK *mtmp)
#else
void free_macro_args(mtmp)
MACRO_STK *mtmp;
#endif
{
  if (mtmp->argl != NULL)
    FREE(mtmp->argl);
  FREE(mtmp);

  return;
}

#ifdef __STDC__
void free_macrodef(MACRODEF *mtmp)
#else
void free_macrodef(mtmp)
MACRODEF *mtmp;
#endif
{
  FREE(mtmp->name);
  FREE(mtmp->body);
  if (mtmp->args != NULL)
    FREE(mtmp->args);
  FREE(mtmp);

  return;
}

#ifdef __STDC__
void setup_macro_args(MACRO_STK *mtmp, TOKEN_LIST *args, TOKEN_LIST *actual)
#else
void setup_macro_args(mtmp, args, actual)
MACRO_STK *mtmp;
TOKEN_LIST *args, *actual;
#endif
{
  static ARGLIST argl[256];
  TOKEN_LIST *ttmp, *atmp;
  int argn = 0;

  /* separate args into an array */
  ttmp = args;
  atmp = actual;
  while (ttmp != NULL) {
    argl[argn].labelnum = 0;
    /* setup formal arg portion */
    argl[argn].def = ttmp;
    while (ttmp->next != NULL) {
      if (ttmp->type == ',')
	break;
      ttmp = ttmp->next;
    }
    ttmp = ttmp->next;

    /* setup actual arg portion */
    if (atmp != NULL) {
      if (atmp->type == ',')
	argl[argn++].act = NULL;
      else {
	argl[argn++].act = atmp;
	/* next should be a comma, or end of arglist */
	if (atmp->next != NULL)
	  atmp = atmp->next;
      }
      atmp = atmp->next;
    } else
      argl[argn++].act = NULL;
  }

  mtmp->argn = argn;
  mtmp->argl = ALLOC0(argn, ARGLIST);
  memcpy(mtmp->argl, argl, argn*sizeof(ARGLIST));

  return;
}

#ifdef __STDC__
void define_macro(TOKEN_LIST *name, TOKEN_LIST *args, TOKEN_LIST *body)
#else
void define_macro(name, args, body)
TOKEN_LIST *name, *args, *body;
#endif
{
  TOKEN_LIST *temp;
  MACRODEF *mtmp = ALLOC0(1, MACRODEF);

#ifdef DEBUG
  fprintf(stderr, "******** defining macro %s\n", name->tok.s);
  fprintf(stderr, "******** body:%s\n", body->tok.s);
#endif

  mtmp->name = name;
  mtmp->body = body;
  mtmp->args = args;

  avl_insert(macrotree, name->tok.s, mtmp);

  return;
}

#ifdef __STDC__
void expand_macro(TOKEN_LIST *name, TOKEN_LIST *args)
#else
void expand_macro(name, args)
TOKEN_LIST *name, *args;
#endif
{
  char *real_body;
  MACRODEF *mtmp;
  MACRO_STK *stktmp;

#ifdef DEBUG
  fprintf(stderr, "******** expanding macro %s\n",name->tok.s);
#endif

  if (!avl_lookup(macrotree, name->tok.s, &mtmp)) {
    errout(E_ERR, "expand_macro", "macro %s doesn't exist, this is not possible.\n", name->tok.s);
    return;
  }

  /* Okay, mtmp contains the macro definition structure...  Let's expand it */

  /* first push cur_ifile back on ifile stack */
  cur_ifile->yychar = yychar;
  cur_ifile->yylval = yylval;
  cur_ifile->next = ifile;
  ifile = cur_ifile;

#ifdef DEBUG
  fprintf(stderr, "******** body:%s\n", mtmp->body->tok.s);
#endif

  stktmp = ALLOC0(1, MACRO_STK);

  setup_macro_args(stktmp, mtmp->args, args);

  stktmp->name = name->tok.s;  /* store away name, it's just going to get thrown away otherwise */
  stktmp->next = macstk;
  macstk = stktmp;

  /* add macro body to input stack */
  real_body = strcpy(ALLOC_STR(strlen(mtmp->body->tok.s)), mtmp->body->tok.s);
  add_input(real_body, F_MACRO, 1, ADD_PUSH);

  /* call next_file() to get input file */
  (void)next_file();

  FREE(name);  /* don't use free_tok because we saved the ->tok.s value in macro stack above */

  return;
}

#ifdef __STDC__
void do_ascii(TOKEN_LIST *str)
#else
void do_ascii(str)
TOKEN_LIST *str;
#endif
{
  char *ctmp = str->tok.s, tstr[50];
  unsigned long temp;

  while (strlen(ctmp) >= 4) {
    temp = *((unsigned long *)ctmp);
    sprintf(tstr, "\n\t.long\t^x%08X", temp);
    add_to_line(tstr);
    ctmp += 4;
  }

  if (*ctmp) {
    int i;

    temp = 0;
    for (i = 0; i < strlen(ctmp); ++i)
      temp |= *ctmp << (i * 8);
    sprintf(tstr, "\n\t.long\t^x%08X", temp);
    add_to_line(tstr);
  }

  free_tok(str);

  return;
}

#ifdef __STDC__
void do_asciz(TOKEN_LIST *str)
#else
void do_asciz(str)
TOKEN_LIST *str;
#endif
{
  char *ctmp = str->tok.s, tstr[50];
  unsigned long temp;

  while (strlen(ctmp) >= 4) {
    temp = *((unsigned long *)ctmp);
    sprintf(tstr, "\n\t.long\t^x%08X", temp);
    add_to_line(tstr);
    ctmp += 4;
  }

  if (*ctmp) {
    /* if we weren't a multiple of 4, then we automatically get a zero
       terminator (because we will pad to longword boundary) */
    int i;

    temp = 0;
    for (i = 0; i < strlen(ctmp); ++i)
      temp |= *ctmp << (i * 8);
    sprintf(tstr, "\n\t.long\t^x%08X", temp);
    add_to_line(tstr);
  } else
    /* add zero termination (and pad to next longword) */
    add_to_line("\n\t.long\t^x00000000");

  free_tok(str);

  return;
}
  
#ifdef __STDC__
void do_psect(TOKEN_LIST *name, TOKEN_LIST *args)
#else
void do_psect(name, args)
TOKEN_LIST *name, *args;
#endif
{
  TOKEN_LIST *ttmp;
  int had_align = 0;

  clear_line();
  add_to_line("\t.psect\t");
  add_to_line(name->tok.s);
  add_to_line(", ");
  
  for(ttmp = args; ttmp!=NULL; ttmp=ttmp->next) {
    if (ttmp->type == NUMBER) {
      if (had_align) {
	errout(E_WARN, "do_psect", "multiple alignment arguments specified in psect definition");
	break;
      }
      add_to_line(tok2str(ttmp));
      had_align = 1;
    }
    if (ttmp->type == SYMBOL) {
      if (strcmp(ttmp->tok.s, "long") == 0) {
	if (had_align) {
	  errout(E_WARN, "do_psect", "multiple alignment arguments specified in psect definition");
	  break;
	}
	add_to_line("^x0000000000000002");
	had_align = 1;
      }
      if (strcmp(ttmp->tok.s, "quad") == 0) {
	if (had_align) {
	  errout(E_WARN, "do_psect", "multiple alignment arguments specified in psect definition");
	  break;
	}
	add_to_line("^x0000000000000003");
	had_align = 1;
      }
    }
  }

  /* if no alignment specified, default to long */
  if (!had_align)
    add_to_line("^x0000000000000002");

  cur_psect = name;
  cur_psect->next = NULL;

  return;
}

void do_save_psect()
{
  TOKEN_LIST *ttmp = copy_expr(cur_psect);

  if (ttmp != NULL) {
    ttmp->next = psect_stk;
    psect_stk = ttmp;
  } else
    errout(E_ERR, "do_save_psect", "current psect not defined.\n");
  
  return;
}

void do_rest_psect()
{
  TOKEN_LIST *ttmp;

  if (psect_stk == NULL) {
    errout(E_ERR, "do_rest_psect", "no psects defined on psect stack.\n");
    return;
  }

  ttmp = psect_stk;
  psect_stk = psect_stk->next;

  clear_line();
  add_to_line("\t.psect\t");
  add_to_line(ttmp->tok.s);

  cur_psect = ttmp;
  cur_psect->next = NULL;

  return;
}

